
/*
 * slotsig_base.h header file
 *
 * Part of the SlogSig project, an attempt to provide a type-safe
 * signals/slots mechanism for the C++ language.
 * This file provides base support for signals and slots.
 *
 * Copyright (C) 2003 Yves Bailly <kafka_AT_kafka-fr_DOT_net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
*/

#ifndef __SLOTSIG_BASES_H__
#define __SLOTSIG_BASES_H__ 1

#include <vector>
#include <map>

namespace SlotSig {

enum SlotSigErrors
{
  NoError = 0,
  NullGlobalFct,
  KnownGlobalFct,
  UnknownGlobalFct,
  DynCastFailed,
  NullInstance,
  KnownSlot,
  UnknownSlot,
  IsEmitting,
  EmitInterrupted,
  NullSignal
} ; // enum SlotSigErrors

const char* error2str(SlotSigErrors error) ;

class SlotBase
{
  public:
    virtual ~SlotBase() ;
} ; // SlotBase

class SlotsSetBase ;

class SignalBase
{
  friend class SlotsSetBase ;

  public:
    SignalBase() ;
    virtual ~SignalBase() ;

  protected:
    virtual void slotsSetDeleted (SlotsSetBase* ssb) = 0 ;

} ; // class SignalBase

class SlotsSetBase
{
    friend class SignalBase ;
    typedef std::map<SignalBase*, unsigned> signals_map_t ;

  public:
    SlotsSetBase() ;
    virtual ~SlotsSetBase() ;

    void disconnectAll() ;

    void connectFrom(SignalBase* sig) ;
    void disconnectFrom(SignalBase* sig) ;
    void signalDeleted(SignalBase* sig) ;

  private:
    signals_map_t signals_map ;
} ; // class SlotsSetBase

template <class _SIGNAL, class _SLOT_SET>
SlotSigErrors connect_const (_SIGNAL* sig, _SLOT_SET* sl_set, const std::string& slot_name)
{
  typename _SIGNAL::MethodTypes<_SLOT_SET>::method_const_t m ;
  if ( get_method(slot_name, m) )
    return sig->connect(sl_set, m) ;
  else
    return SlotSig::UnknownSlot ;
}

template <class _SIGNAL, class _SLOT_SET>
SlotSigErrors connect (_SIGNAL* sig, _SLOT_SET* sl_set, const std::string& slot_name)
{
  typename _SIGNAL::MethodTypes<_SLOT_SET>::method_t m ;
  if ( get_method(slot_name, m) )
    return sig->connect(sl_set, m) ;
  else
    return SlotSig::UnknownSlot ;
}

#define SLOTSIG_DECLARE_SLOTS_INCLASS(clas, ret_type, params) \
  namespace SlotSig { \
  bool get_method(const std::string& name, ret_type (clas::*& method) params) ; \
  }

#define SLOTSIG_DECLARE_CONST_SLOTS_INCLASS(clas, ret_type, params) \
  namespace SlotSig { \
  bool get_method(const std::string& name, ret_type (clas::*& method) params const) ; \
  }

#define SLOTSIG_GET_SLOT_INCLASS(clas, ret_type, params) \
  namespace SlotSig { \
  bool get_method(const std::string& name, ret_type (clas::*& method) params) \
  { typedef clas InClass ; \
    typedef ret_type (clas::* MethodType) params ; \
    typedef std::map<std::string, MethodType> MethodsMapType ; \
    static bool init_done = false ; \
    static MethodsMapType methods_map ; \
    if ( ! init_done ) \
    {

#define SLOTSIG_GET_CONST_SLOT_INCLASS(clas, ret_type, params) \
  namespace SlotSig { \
  bool get_method(const std::string& name, ret_type (clas::*& method) params const) \
  { typedef clas InClass ; \
    typedef ret_type (clas::* MethodType) params const ; \
    typedef std::map<std::string, MethodType> MethodsMapType ; \
    static bool init_done = false ; \
    static MethodsMapType methods_map ; \
    if ( ! init_done ) \
    {

#define SLOTSIG_SLOT_INCLASS(meth) \
      methods_map.insert(MethodsMapType::value_type(#meth, &InClass::meth)) ;

#define SLOTSIG_GET_SLOT_INCLASS_END \
      init_done = true ; \
    } \
    MethodsMapType::const_iterator i = methods_map.find(name) ; \
    if ( i == methods_map.end() ) \
      return false ; \
    method = i->second ; \
    return true ; \
  } }

} // namespace SlotSig

#endif // __SLOTSIG_BASES_H__
