
/*
 * slotsig_storage_policies.h header file
 *
 * Part of the SlogSig project, an attempt to provide a type-safe
 * signals/slots mechanism for the C++ language.
 * This file provides storage policies to customize behaviours of signals/slots.
 *
 * Copyright (C) 2003 Yves Bailly <kafka_AT_kafka-fr_DOT_net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
*/

#include <set>
#include <vector>
#include <algorithm>

namespace SlotSig
{

/*
  Here are defined storage classes, which will be used to store pointers to
  slots connected to a signal.

  You can provide your own storage class, for example :
    template <class _TYPE, class _EQUAL_SLOT>
    class MyStorage
    { ... } ;
    SlotSig::Sig1<void, int, MyStorage> my_signal ;
  (requisites for storage classes are described below)

  When you use the default declaration :
    SlotSig::Sig1<void, int> my_signal ;
  ...by default the SimpleStorage is used.

  If you provide you own storage class, it must follow these rules :
  - it must accept two template parameters, the first one being the stored
    type (it will actually always be pointers to this type), the second one
    being a functor allowing to compare two slots for equality ;
    let's name "_TYPE" the first, "_EQUAL_SLOT" the second ;
  - it must define two internal types, "iterator" and "const_iterator",
    which are iterators used to parse the stored objects ;
  - it must provide methods "begin()" and "end()" to get appropriate
    iterators, just like standards containers ;
  - it must provide a "bool insert(_TYPE v)", used to store an object ;
    this method should return "true" on success, "false" on failure ;
    it's up to you if it should allow connecting the same slot to the
    same signal more than once ;
  - it must provide a "iterator erase(iterator i)" method, used to remove
    an item (just like e.g. std::vector) ; the returned iterator should
    refers to the item following i, or end() if there's none ;
  - it must provide a "clear()" method, which will remove all stored
    items. It is not required that the allocated storage memory is
    actually freed.

  It's ok to allow to store null pointers. In any case, you should NEVER
  delete stored pointers, and NEVER try to duplicate their content. The
  storage class stores the pointers, it doesn't own them.
  Also you don't have to check if a pointer is stored twice (or two
  pointers refering to the same slot), the SlotN<> classes already take care
  of this.

  A note on _EQUAL_SLOT. It can be used in two ways, assuming you have
  two slots "s1" and "s2" (both are pointers, remember) :
  1- use it like an unary function :
      _EQUAL_SLOT eq(s1) ;
      eq(s2) ;
    ...will return "true" if s1 and s2 point to the same slot, "false"
    otherwise ; this can be used in algorithms such as std::find_if() ;
  2- use it like a binary function :
    _EQUAL_SLOT()(s1, s2) ;
    ...will return "true" if s1 and s2 point to the same slot, "false"
    otherwise.
  Keep in mind that s1 and s2 may be different, considering their "numerical"
  pointer values, but contain the same slot !
*/

template <typename _TYPE/*, class _EQUAL_TYPE = std::equal_to<_TYPE*>*/>
class SimpleStorage
{
  public:
    typedef typename std::vector<_TYPE> storage_t ;
    typedef typename storage_t::const_iterator const_iterator ;
    typedef typename storage_t::iterator iterator ;

    SimpleStorage() :
      storage(),
      nb_stored(0)
    {
      free_place = end() ;
    }

    ~SimpleStorage()
    { }

    const_iterator begin() const
    { return storage.begin() ; }
    iterator begin()
    { return storage.begin() ; }
    const_iterator end() const
    { return storage.end() ; }
    iterator end()
    { return storage.end() ; }

    size_t capacity() const
    { return storage.size() ; }

    unsigned stored() const
    { return nb_stored ; }

    void clear()
    {
      std::fill(begin(), end(), (_TYPE)0) ;
      nb_stored = 0 ;
      free_place = begin() ;
    }

    template <class _C, typename _METHOD_TYPE, class _TRUE_TYPE>
    iterator find(_C* in_class, _METHOD_TYPE m, _TRUE_TYPE* dummy)
    {
      const iterator e = end() ;
      iterator i = begin() ;
      while ( i != e )
      {
        if ( (*i) != 0 )
        {
          _TRUE_TYPE* tt = static_cast<_TRUE_TYPE*>(*i) ;
          if ( tt != 0 )
            if ( tt->matches(in_class, m) )
              return i ;
        }
        ++i ;
      }
      return e ;
    }

    iterator find_equal(_TYPE v)
    {
      const iterator e = end() ;
      iterator i = begin() ;
      while ( i != e )
      {
        if ( (*i) != 0 )
          if ( (*i)->equals(v) )
            return i ;
        ++i ;
      }
      return e ;
    }

    bool insert(_TYPE v)
    {
      if ( nb_stored == storage.size() )
      {
        storage.resize(nb_stored+(nb_stored>>1)+1, 0) ;
        free_place = end() ;
      }
      iterator place = find_free_place() ;
      const iterator e = end() ;
      if ( place == e )
      {
        storage.push_back(v) ;
      }
      else
      {
        *place = v ;
        free_place = ++place ;
        if ( (free_place != e) && ((*free_place) != 0) )
          free_place = e ;
      }
      ++nb_stored ;
      return true ;
    }

    iterator erase(iterator iter)
    {
      if ( iter == end() )
        return iter ;
      *iter = 0 ;
      free_place = iter ;
      --nb_stored ;
      return iter+1 ;
    }

  protected:
    iterator find_free_place()
    {
      const iterator e = end() ;
      if ( free_place == e )
      {
        free_place = begin() ;
        while ( (free_place != e) && (*free_place != 0) )
          ++free_place ;
      }
      return free_place ;
    }

  private:
    storage_t storage ;
    unsigned nb_stored ;
    iterator free_place ;
} ; // class SimpleStorage<>

} // namespace SlotSig
